package animation

import "../../tlr"
import "../../tlr/gpu"
import "core:image/qoi"
import "core:log"
import "core:time"

Vertex :: struct{
	position: [2]f32,
	uv: [2]f32,
}

vertices := []Vertex{
	{{ 0.5, 0.5 }, { 1, 1 }},
	{{ 0.5, -0.5 }, { 1, 0 }},
	{{ -0.5, 0.5 }, { 0, 1 }},
	{{ -0.5, -0.5 }, { 0, 0 }},
}

indicies := []u32{
	0, 1, 2,
	1, 3, 2,
}

FRAMES :: 8
FRAME_TIME :: 1.0/10.0

main :: proc() {
	context.logger = log.create_console_logger()
	defer log.destroy_console_logger(context.logger)

	when gpu.BACKEND == .Vulkan {
		gpu.vk_set_hints(gpu.Vk_Hints{
			use_validation_layers = true,
		})
	}

	tlr.init("Animation Example", 500, 500)
	gpu.init()
	defer {
		gpu.cleanup()
		tlr.cleanup()
	}

	vert_shader, vert_ok := gpu.make_shader("./examples/animation/animation.vert.spv")
	frag_shader, frag_ok := gpu.make_shader("./examples/animation/animation.frag.spv")
	assert(vert_ok && frag_ok)

	run_image, run_image_err := qoi.load_from_file("./examples/animation/run.qoi")
	assert(run_image_err == nil)
	defer qoi.destroy(run_image)

	frame_width, frame_height := run_image.width, run_image.height/FRAMES
	bytes := frame_width*frame_height*4

	texture, _ := gpu.make_texture(gpu.Texture_Info{
		size = [2]int{ frame_width, frame_height, },
		array_len = FRAMES,
		filter_mode = .Near,
	})

	for i in 0..<FRAMES {
		start := i*bytes
		gpu.upload_texture(texture, run_image.pixels.buf[start:start + bytes], array_layer = i)
	}

	buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(f32),
	})
	frame_index := f32(0)
	gpu.upload_buffer(buffer, &frame_index)

	texture_descriptor_info := gpu.Descriptor_Info{
		binding = 0,
		type = .Texture,
		stages = { .Fragment },
	}
	buffer_descriptor_info := gpu.Descriptor_Info{
		binding = 1,
		type = .Uniform_Buffer,
		stages = { .Fragment },
	}
	descriptors, _ := gpu.make_descriptors({ texture_descriptor_info, buffer_descriptor_info })
	gpu.update_descriptors_texture(descriptors, 0, texture)
	gpu.update_descriptors_buffer(descriptors, 1, buffer)

	vertex_binding := gpu.Vertex_Binding{
		binding = 0,
		stride = size_of(Vertex),
		attributes = {
			{
				location = 0,
				format = .Vec2_Float32,
				offset = offset_of(Vertex, position),
			},
			{
				location = 1,
				format = .Vec2_Float32,
				offset = offset_of(Vertex, uv),
			},
		},
	}
	pipeline_info := gpu.Pipeline_Info{
		vertex = vert_shader,
		fragment = frag_shader,

		vertex_bindings = { vertex_binding, },
		descriptors = { descriptors },
	}
	pipeline, _ := gpu.make_pipeline(pipeline_info)

	vert_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(Vertex)*len(vertices),
	})
	gpu.upload_buffer(vert_buffer, vertices)

	ind_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(u32)*len(indicies),
	})
	gpu.upload_buffer(ind_buffer, indicies)

	accumulator: f32
	clock: time.Stopwatch
	time.stopwatch_start(&clock)
	for !tlr.window_should_close() {
		delta := cast(f32)time.duration_seconds(time.stopwatch_duration(clock))
		time.stopwatch_reset(&clock)
		time.stopwatch_start(&clock)

		for accumulator += delta; accumulator >= FRAME_TIME; accumulator -= FRAME_TIME {
			using gpu
			frame_index += 1
			if frame_index == FRAMES {
				frame_index = 0
			}
			gpu.upload_buffer(buffer, &frame_index)

			begin_pass(Pass_Info{
				clear_color = [4]f32{ 0, 0.5, 0.5, 1 },
			})
			bind_pipeline(pipeline)
			bind_vertex_buffer(vert_buffer)
			bind_index_buffer(ind_buffer, .UInt32)
			bind_descriptors(descriptors)
			draw_elems_indexed(len(indicies))
			end_pass()
		}
	}
}
