package cube

import "../../tlr"
import "../../tlr/gpu"
import "core:image/qoi"
import glm "core:math/linalg/glsl"
import "core:log"
import "core:time"

Vertex :: struct{
	position: [3]f32,
	uv: [2]f32,
}

vertices := []Vertex{
	{{ -1, 1, 1, }, { 0, 1, }},
	{{ 1, -1, 1, }, { 1, 0, }},
	{{ -1, -1, 1, }, { 0, 0, }},

	{{ 1, -1, 1, }, { 1, 0, }},
	{{ -1, 1, 1, }, { 0, 1, }},
	{{ 1, 1, 1, }, { 1, 1, }},

	{{ 1, -1, -1, }, { 1, 0, }},
	{{ -1, 1, -1, }, { 0, 1, }},
	{{ -1, -1, -1, }, { 0, 0, }},

	{{ -1, 1, -1, }, { 0, 1, }},
	{{ 1, -1, -1, }, { 1, 0, }},
	{{ 1, 1, -1, }, { 1, 1, }},

	{{ -1, -1, -1, }, { 0, 0, }},
	{{ -1, -1, 1, }, { 0, 1, }},
	{{ 1, -1, 1, }, { 1, 1, }},

	{{ 1, -1, -1, }, { 1, 0, }},
	{{ -1, -1, -1, }, { 0, 0, }},
	{{ 1, -1, 1, }, { 1, 1, }},

	{{ 1, 1, 1, }, { 1, 1, }},
	{{ -1, 1, 1, }, { 0, 1, }},
	{{ -1, 1, -1, }, { 0, 0, }},

	{{ 1, 1, 1, }, { 1, 1, }},
	{{ -1, 1, -1, }, { 0, 0, }},
	{{ 1, 1, -1, }, { 1, 0, }},

	{{ 1, 1, 1, }, { 1, 1, }},
	{{ 1, 1, -1, }, { 1, 0, }},
	{{ 1, -1, -1, }, { 0, 0, }},

	{{ 1, 1, 1, }, { 1, 1, }},
	{{ 1, -1, -1, }, { 0, 0, }},
	{{ 1, -1, 1, }, { 0, 1, }},

	{{ -1, 1, -1, }, { 1, 0, }},
	{{ -1, 1, 1, }, { 1, 1, }},
	{{ -1, -1, -1, }, { 0, 0, }},

	{{ -1, -1, -1, }, { 0, 0, }},
	{{ -1, 1, 1, }, { 1, 1, }},
	{{ -1, -1, 1, }, { 0, 1, }},
}

UBO :: struct{
	model: glm.mat4,
	view: glm.mat4,
	projection: glm.mat4,
}

FRAME_TIME :: 1.0/60.0

main :: proc() {
	context.logger = log.create_console_logger()
	defer log.destroy_console_logger(context.logger)

	when gpu.BACKEND == .Vulkan {
		gpu.vk_set_hints(gpu.Vk_Hints{
			use_validation_layers = true,
		})
	}

	tlr.init("Cube Example", 500, 500)
	gpu.init()
	defer {
		gpu.cleanup()
		tlr.cleanup()
	}

	vert_shader, vert_ok := gpu.make_shader("./examples/cube/cube.vert.spv")
	frag_shader, frag_ok := gpu.make_shader("./examples/cube/cube.frag.spv")
	assert(vert_ok && frag_ok)

	depth_texture, _ := gpu.make_depth_texture(gpu.Depth_Texture_Info{
		size = .Swapchain,
		filter_mode = .Near,
	})

	birds_image, birds_image_err := qoi.load_from_file("./examples/cube/birds.qoi", qoi.Options{ .alpha_add_if_missing })
	assert(birds_image_err == nil)
	defer qoi.destroy(birds_image)

	fovy := glm.radians(f32(45))
	ubo := UBO{
		model = glm.identity(glm.mat4),
		view = glm.mat4LookAt({ 5, 3, 0, }, { 0, 0, 0, }, { 0, -1, 0, }),
		projection = glm.mat4Perspective(fovy, 1, 0.5, 10),
	}
	ubo_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(UBO),
	})
	gpu.upload_buffer(ubo_buffer, &ubo)

	texture, _ := gpu.make_texture(gpu.Texture_Info{
		writable = false,
		size = [2]int{ birds_image.width, birds_image.height },
	})
	gpu.upload_texture(texture, birds_image.pixels.buf[:])

	ubo_descriptor_info := gpu.Descriptor_Info{
		binding = 0,
		count = 1,
		type = .Uniform_Buffer,
		stages = { .Vertex },
	}
	texture_descriptor_info := gpu.Descriptor_Info{
		binding = 1,
		count = 1,
		type = .Texture,
		stages = { .Fragment },
	}
	descriptors, _ := gpu.make_descriptors({ ubo_descriptor_info, texture_descriptor_info })
	gpu.update_descriptors_buffer(descriptors, 0, ubo_buffer)
	gpu.update_descriptors_texture(descriptors, 1, texture)

	vertex_binding := gpu.Vertex_Binding{
		binding = 0,
		stride = size_of(Vertex),
		attributes = {
			{
				location = 0,
				format = .Vec3_Float32,
				offset = offset_of(Vertex, position),
			},
			{
				location = 1,
				format = .Vec2_Float32,
				offset = offset_of(Vertex, uv),
			},
		},
	}
	pipeline_info := gpu.Pipeline_Info{
		vertex = vert_shader,
		fragment = frag_shader,

		vertex_bindings = { vertex_binding, },
		descriptors = { descriptors },

		depth_enable = .Enabled,
		depth_write = true,
		depth_op = .Less,
	}
	pipeline, _ := gpu.make_pipeline(pipeline_info)

	vert_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(Vertex)*len(vertices),
	})
	gpu.upload_buffer(vert_buffer, vertices)

	accumulator: f32
	clock: time.Stopwatch
	time.stopwatch_start(&clock)
	for !tlr.window_should_close() {
		delta := cast(f32)time.duration_seconds(time.stopwatch_duration(clock))
		time.stopwatch_reset(&clock)
		time.stopwatch_start(&clock)

		for accumulator += delta; accumulator >= FRAME_TIME; accumulator -= FRAME_TIME {
			using gpu
			ubo.model *= glm.mat4Rotate({ 0, 1, 0, }, 0.03)
			upload_buffer(ubo_buffer, &ubo)

			begin_pass(Pass_Info{
				clear_color = [4]f32{ 0, 0.5, 0.5, 1 },
				depth_stencil_info = gpu.Depth_Stencil_Info{
					texture = depth_texture,
					depth_clear = 1,
				},
			})
			bind_pipeline(pipeline)
			bind_vertex_buffer(vert_buffer)
			bind_descriptors(descriptors)
			draw_elems(len(vertices))
			end_pass()
		}
	}
}
