package model

import "../../tlr"
import "../../tlr/gpu"
import glm "core:math/linalg/glsl"
import "core:image/png"
import "core:log"
import "core:time"
import "core:os"
import "core:strings"
import "core:strconv"

Vertex :: struct{
	position: [3]f32,
	uv: [2]f32,
}

UBO :: struct{
	model: glm.mat4,
	view: glm.mat4,
	projection: glm.mat4,
}

FRAME_TIME :: 1.0/60.0

main :: proc() {
	context.logger = log.create_console_logger()
	defer log.destroy_console_logger(context.logger)

	when gpu.BACKEND == .Vulkan {
		gpu.vk_set_hints(gpu.Vk_Hints{
			use_validation_layers = true,
		})
	}

	tlr.init("Model Example", 500, 500)
	gpu.init()
	defer {
		gpu.cleanup()
		tlr.cleanup()
	}

	vert_shader, vert_ok := gpu.make_shader("./examples/model/model.vert.spv")
	frag_shader, frag_ok := gpu.make_shader("./examples/model/model.frag.spv")
	assert(vert_ok && frag_ok)

	vert_buffer, count := load_model("./examples/model/viking_room.obj")
	viking_room_png, _ := png.load_from_file("./examples/model/viking_room.png", { .alpha_add_if_missing })
	defer png.destroy(viking_room_png)

	texture, _ := gpu.make_texture(gpu.Texture_Info{
		size = [2]int{ viking_room_png.width, viking_room_png.height, },
		mip_levels = 3,
	})
	gpu.upload_texture(texture, viking_room_png.pixels.buf[:])
	gpu.gen_texture_mip_chain(texture)

	fovy := glm.radians(f32(45))
	ubo := UBO{
		model = glm.identity(glm.mat4),
		view = glm.mat4LookAt({ 2, 0, 2, }, { 0, 0, 0, }, { 0, 0, -1, }),
		projection = glm.mat4Perspective(fovy, 1, 0.1, 100),
	}
	ubo_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		usage = .Uniform,
		size = size_of(UBO),
	})
	gpu.upload_buffer(ubo_buffer, &ubo)

	ubo_descriptor_info := gpu.Descriptor_Info{
		binding = 0,
		count = 1,
		type = .Uniform_Buffer,
		stages = { .Vertex },
	}
	texture_descriptor_info := gpu.Descriptor_Info{
		binding = 1,
		count = 1,
		type = .Texture,
		stages = { .Fragment },
	}
	descriptors, _ := gpu.make_descriptors({ ubo_descriptor_info, texture_descriptor_info })
	gpu.update_descriptors_buffer(descriptors, 0, ubo_buffer)
	gpu.update_descriptors_texture(descriptors, 1, texture)

	depth_texture, _ := gpu.make_depth_texture(gpu.Depth_Texture_Info{
		size = .Swapchain,
	})

	vertex_binding := gpu.Vertex_Binding{
		binding = 0,
		stride = size_of(Vertex),
		attributes = {
			{
				location = 0,
				format = .Vec3_Float32,
				offset = offset_of(Vertex, position),
			},
			{
				location = 1,
				format = .Vec2_Float32,
				offset = offset_of(Vertex, uv),
			},
		},
	}
	pipeline_info := gpu.Pipeline_Info{
		vertex = vert_shader,
		fragment = frag_shader,

		vertex_bindings = { vertex_binding },
		descriptors = { descriptors },

		cull_mode = .None,

		depth_enable = .Enabled,
		depth_write = true,
		depth_op = .Less,
	}
	pipeline, _ := gpu.make_pipeline(pipeline_info)

	accumulator: f32
	clock: time.Stopwatch
	time.stopwatch_start(&clock)
	for !tlr.window_should_close() {
		delta := cast(f32)time.duration_seconds(time.stopwatch_duration(clock))
		time.stopwatch_reset(&clock)
		time.stopwatch_start(&clock)

		for accumulator += delta; accumulator >= FRAME_TIME; accumulator -= FRAME_TIME {
			using gpu
			ubo.model *= glm.mat4Rotate({ 0, 0, 1, }, 0.003)
			upload_buffer(ubo_buffer, &ubo)

			begin_pass(Pass_Info{
				clear_color = [4]f32{ 0, 0.5, 0.5, 1 },
				depth_stencil_info = gpu.Depth_Stencil_Info{
					texture = depth_texture,
					depth_clear = 1,
				},
			})
			bind_pipeline(pipeline)
			bind_descriptors(descriptors)
			bind_vertex_buffer(vert_buffer)
			draw_elems(count)
			end_pass()
		}
	}
}

load_model :: proc(filename: string) -> (buffer: gpu.Buffer, count: int) {
	defer free_all(context.temp_allocator)

	data, ok := os.read_entire_file(filename, allocator = context.temp_allocator)
	assert(ok)

	positions := make([dynamic][3]f32, allocator = context.temp_allocator)
	uvs := make([dynamic][2]f32, allocator = context.temp_allocator)
	indices := make([dynamic][2]int, allocator = context.temp_allocator)

	line_iterator := string(data)
	for l in strings.split_lines_iterator(&line_iterator) {
		if len(l) == 0 || l[0] == '#' {
			continue
		}

		tokens := strings.fields(l, allocator = context.temp_allocator)

		switch tokens[0] {
		case "v":
			position: [3]f32
			position.x, _ = strconv.parse_f32(tokens[1])
			position.y, _ = strconv.parse_f32(tokens[2])
			position.z, _ = strconv.parse_f32(tokens[3])
			append(&positions, position)

		case "vt":
			uv: [2]f32
			uv.x, _ = strconv.parse_f32(tokens[1])
			uv.y, _ = strconv.parse_f32(tokens[2])
			uv.y = 1 - uv.y
			append(&uvs, uv)

		case "f":
			for t in tokens[1:] {
				vertex := strings.split(t, "/", context.temp_allocator)
				index: [2]int
				index[0], _ = strconv.parse_int(vertex[0])
				index[1], _ = strconv.parse_int(vertex[1])
				append(&indices, index)
			}
		}
	}

	verticies := make([dynamic]Vertex, allocator = context.temp_allocator)
	for i in indices {
		pos := positions[i[0] - 1]
		uv := uvs[i[1] - 1]
		append(&verticies, Vertex{
			position = pos,
			uv = uv,
		})
	}

	buffer, _ = gpu.make_buffer(gpu.Buffer_Info{
		usage = .Vertex,
		size = size_of(Vertex)*len(verticies),
	})
	gpu.upload_buffer(buffer, verticies[:])
	return buffer, len(verticies)
}
