package cube

import "../../tlr"
import "../../tlr/gpu"
import glm "core:math/linalg/glsl"
import "core:log"
import "core:time"

Cube_Vertex :: struct{
	position: [3]f32,
	uv: [2]f32,
}

Sqaure_Vertex :: struct{
	position: [2]f32,
	color: [4]f32,
}

cube_vertices := []Cube_Vertex{
	{{ -1, 1, 1, }, { 0, 1, }},
	{{ 1, -1, 1, }, { 1, 0, }},
	{{ -1, -1, 1, }, { 0, 0, }},

	{{ 1, -1, 1, }, { 1, 0, }},
	{{ -1, 1, 1, }, { 0, 1, }},
	{{ 1, 1, 1, }, { 1, 1, }},

	{{ 1, -1, -1, }, { 1, 0, }},
	{{ -1, 1, -1, }, { 0, 1, }},
	{{ -1, -1, -1, }, { 0, 0, }},

	{{ -1, 1, -1, }, { 0, 1, }},
	{{ 1, -1, -1, }, { 1, 0, }},
	{{ 1, 1, -1, }, { 1, 1, }},

	{{ -1, -1, -1, }, { 0, 0, }},
	{{ -1, -1, 1, }, { 0, 1, }},
	{{ 1, -1, 1, }, { 1, 1, }},

	{{ 1, -1, -1, }, { 1, 0, }},
	{{ -1, -1, -1, }, { 0, 0, }},
	{{ 1, -1, 1, }, { 1, 1, }},

	{{ 1, 1, 1, }, { 1, 1, }},
	{{ -1, 1, 1, }, { 0, 1, }},
	{{ -1, 1, -1, }, { 0, 0, }},

	{{ 1, 1, 1, }, { 1, 1, }},
	{{ -1, 1, -1, }, { 0, 0, }},
	{{ 1, 1, -1, }, { 1, 0, }},

	{{ 1, 1, 1, }, { 1, 1, }},
	{{ 1, 1, -1, }, { 1, 0, }},
	{{ 1, -1, -1, }, { 0, 0, }},

	{{ 1, 1, 1, }, { 1, 1, }},
	{{ 1, -1, -1, }, { 0, 0, }},
	{{ 1, -1, 1, }, { 0, 1, }},

	{{ -1, 1, -1, }, { 1, 0, }},
	{{ -1, 1, 1, }, { 1, 1, }},
	{{ -1, -1, -1, }, { 0, 0, }},

	{{ -1, -1, -1, }, { 0, 0, }},
	{{ -1, 1, 1, }, { 1, 1, }},
	{{ -1, -1, 1, }, { 0, 1, }},
}

square_vertices := []Sqaure_Vertex{
	{{ 0.5, 0.5 }, { 1, 0, 0, 1 }},
	{{ 0.5, -0.5 }, { 0, 1, 0, 1 }},
	{{ -0.5, 0.5 }, { 0, 0, 1, 1 }},
	{{ -0.5, -0.5 }, { 0, 0.5, 0.5, 1 }},
}

square_indicies := []u32{
	0, 1, 2,
	1, 3, 2,
}

UBO :: struct{
	model: glm.mat4,
	view: glm.mat4,
	projection: glm.mat4,
}

FRAME_TIME :: 1.0/60.0

main :: proc() {
	context.logger = log.create_console_logger()
	defer log.destroy_console_logger(context.logger)

	when gpu.BACKEND == .Vulkan {
		gpu.vk_set_hints(gpu.Vk_Hints{
			use_validation_layers = true,
		})
	}

	tlr.init("Render Target Example", 500, 500)
	gpu.init()
	defer {
		gpu.cleanup()
		tlr.cleanup()
	}

	depth_texture, _ := gpu.make_depth_texture(gpu.Depth_Texture_Info{
		size = .Swapchain,
		filter_mode = .Near,
	})

	square_render_texture, _ := gpu.make_render_texture(gpu.Render_Texture_Info{ size = .Swapchain })
	cube_render_texture, _ := gpu.make_render_texture(gpu.Render_Texture_Info{ size = .Swapchain })

	fovy := glm.radians(f32(45))
	ubo := UBO{
		model = glm.identity(glm.mat4),
		view = glm.mat4LookAt({ 5, 3, 0, }, { 0, 0, 0, }, { 0, -1, 0, }),
		projection = glm.mat4Perspective(fovy, 1, 0.5, 10),
	}

	ubo_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(UBO),
	})
	gpu.upload_buffer(ubo_buffer, &ubo)

	cube_pipeline, cube_descriptors := make_cube_pipeline(ubo_buffer)
	square_pipeline := make_square_pipeline()

	cube_vert_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(Cube_Vertex)*len(cube_vertices),
	})
	gpu.upload_buffer(cube_vert_buffer, cube_vertices)

	square_vert_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(Sqaure_Vertex)*len(square_vertices),
	})
	gpu.upload_buffer(square_vert_buffer, square_vertices)

	square_ind_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(u32)*len(square_indicies),
	})
	gpu.upload_buffer(square_ind_buffer, square_indicies)

	accumulator: f32
	clock: time.Stopwatch
	time.stopwatch_start(&clock)
	for !tlr.window_should_close() {
		delta := cast(f32)time.duration_seconds(time.stopwatch_duration(clock))
		time.stopwatch_reset(&clock)
		time.stopwatch_start(&clock)

		for accumulator += delta; accumulator >= FRAME_TIME; accumulator -= FRAME_TIME {
			using gpu
			ubo.model *= glm.mat4Rotate({ 0, 1, 0, }, 0.03)
			upload_buffer(ubo_buffer, &ubo)

			begin_pass(Pass_Info{
				render_target = square_render_texture,
				clear_color = [4]f32{ 0.5, 0, 0.5, 1 },
			})
			bind_pipeline(square_pipeline)
			bind_vertex_buffer(square_vert_buffer)
			bind_index_buffer(square_ind_buffer, .UInt32)
			draw_elems_indexed(len(square_indicies))
			end_pass()

			begin_pass(Pass_Info{
				render_target = cube_render_texture,
				clear_color = [4]f32{ 0.5, 0.5, 0, 1 },
				depth_stencil_info = gpu.Depth_Stencil_Info{
					texture = depth_texture,
					depth_clear = 1,
				},
			})
			gpu.update_descriptors_texture(cube_descriptors, 1, square_render_texture)
			bind_pipeline(cube_pipeline)
			bind_vertex_buffer(cube_vert_buffer)
			bind_descriptors(cube_descriptors)
			draw_elems(len(cube_vertices))
			end_pass()

			begin_pass(Pass_Info{
				clear_color = [4]f32{ 0, 0.5, 0.5, 1 },
				depth_stencil_info = gpu.Depth_Stencil_Info{
					texture = depth_texture,
					depth_clear = 1,
				},
			})
			gpu.update_descriptors_texture(cube_descriptors, 1, cube_render_texture)
			bind_pipeline(cube_pipeline)
			bind_vertex_buffer(cube_vert_buffer)
			bind_descriptors(cube_descriptors)
			draw_elems(len(cube_vertices))
			end_pass()
		}
	}
}

make_cube_pipeline :: proc(ubo_buffer: gpu.Buffer) -> (gpu.Pipeline, gpu.Descriptors) {
	vert_shader, vert_ok := gpu.make_shader("./examples/render-target/render_target_cube.vert.spv")
	frag_shader, frag_ok := gpu.make_shader("./examples/render-target/render_target_cube.frag.spv")
	assert(vert_ok && frag_ok)

	ubo_descriptor_info := gpu.Descriptor_Info{
		binding = 0,
		count = 1,
		type = .Uniform_Buffer,
		stages = { .Vertex },
	}
	texture_descriptor_info := gpu.Descriptor_Info{
		binding = 1,
		count = 1,
		type = .Texture,
		stages = { .Fragment },
	}
	descriptors, _ := gpu.make_descriptors({ ubo_descriptor_info, texture_descriptor_info })
	gpu.update_descriptors_buffer(descriptors, 0, ubo_buffer)

	vertex_binding := gpu.Vertex_Binding{
		binding = 0,
		stride = size_of(Cube_Vertex),
		attributes = {
			{
				location = 0,
				format = .Vec3_Float32,
				offset = offset_of(Cube_Vertex, position),
			},
			{
				location = 1,
				format = .Vec2_Float32,
				offset = offset_of(Cube_Vertex, uv),
			},
		},
	}
	pipeline_info := gpu.Pipeline_Info{
		vertex = vert_shader,
		fragment = frag_shader,

		vertex_bindings = { vertex_binding, },
		descriptors = { descriptors },

		depth_enable = .Enabled,
		depth_write = true,
		depth_op = .Less,
	}
	pipeline, _ := gpu.make_pipeline(pipeline_info)
	return pipeline, descriptors
}

make_square_pipeline :: proc() -> gpu.Pipeline {
	vert_shader, vert_ok := gpu.make_shader("./examples/render-target/render_target_square.vert.spv")
	frag_shader, frag_ok := gpu.make_shader("./examples/render-target/render_target_square.frag.spv")
	assert(vert_ok && frag_ok)

	vertex_binding := gpu.Vertex_Binding{
		binding = 0,
		stride = size_of(Sqaure_Vertex),
		attributes = {
			{
				location = 0,
				format = .Vec2_Float32,
				offset = offset_of(Sqaure_Vertex, position),
			},
			{
				location = 1,
				format = .Vec4_Float32,
				offset = offset_of(Sqaure_Vertex, color),
			},
		},
	}
	pipeline_info := gpu.Pipeline_Info{
		vertex = vert_shader,
		fragment = frag_shader,

		vertex_bindings = { vertex_binding, },
	}
	pipeline, _ := gpu.make_pipeline(pipeline_info)
	return pipeline
}
